from __future__ import annotations
from enum import Enum
from lalalang.lexer.token import TokenType


class ExpressionPrecedence(Enum):
    """
    These are the operator of la la lang, we assing them to a
    value in order to denote precedence
    """

    def __repr__(self):
        return "ExpressionPrecedence(%s)" % self.value

    def __str__(self):
        return str(self.value)

    def __lt__(self, other: ExpressionPrecedence):
        return self.value - other.value < 0

    LOWEST = 0
    EQUALS = 1
    LESS_GREATER = 2
    SUM = 3
    PRODUCT = 4
    PREFIX = 5
    CALL = 6


PRECEDENCES: dict[TokenType, ExpressionPrecedence] = {
    TokenType.EQ: ExpressionPrecedence.EQUALS,
    TokenType.NOT_EQ: ExpressionPrecedence.EQUALS,
    TokenType.LT: ExpressionPrecedence.LESS_GREATER,
    TokenType.GT: ExpressionPrecedence.LESS_GREATER,
    TokenType.PLUS: ExpressionPrecedence.SUM,
    TokenType.MINUS: ExpressionPrecedence.SUM,
    TokenType.SLASH: ExpressionPrecedence.PRODUCT,
    TokenType.ASTERISK: ExpressionPrecedence.PRODUCT,
    TokenType.LPAREN: ExpressionPrecedence.CALL,
}



SOURCE_CODE = """
    let pi = 314;
    fun area(ratio) {
        pi * ratio * ratio + 0;
    }

    13 / 3 > 16;

    if(10 - 8 < 16) {
        return true;
    } else {
        return false;
    }

    10 == 10;
    10 != 9;
"""

DELIMETERS = "(){},;"

ILLEGAL = "~`^"

OPERATORS = "+=-!*/<>"

TWO_CHARACTER_SYMBOLS = "! == != ="

IDENTIFIERS = "someone in the crowd"

NUMBERS = "314 217 161"

KEYWORDS = "let fun true false if else return"



from lalalang.lexer.token import TokenType, Token, lookup_identifier


class Lexer:
    """
    This class recieves source code as input and returns
    the tokens that represent it
    """

    def __init__(self, source: str):
        self.source: str = source
        self.position: int = 0
        self.peek_position: int = 0
        self.char: str = ""

        self._read_char()

    def next_token(self) -> Token:
        """
        This is the main method that is traversing the
        source code and generating the tokens, this was
        made for being used multiple times
        """
        self._skip_whitespace()
        new_token: Token = Token.empty()
        char: str = self.char

        # One character symbols
        if self.char == ";":
            new_token = Token(TokenType.SEMICOLON, self.char)
        elif self.char == "(":
            new_token = Token(TokenType.LPAREN, self.char)
        elif self.char == ")":
            new_token = Token(TokenType.RPAREN, self.char)
        elif self.char == ",":
            new_token = Token(TokenType.COMMA, self.char)
        elif self.char == "+":
            new_token = Token(TokenType.PLUS, self.char)
        elif self.char == "{":
            new_token = Token(TokenType.LBRACE, self.char)
        elif self.char == "}":
            new_token = Token(TokenType.RBRACE, self.char)
        elif self.char == "-":
            new_token = Token(TokenType.MINUS, self.char)
        elif self.char == "*":
            new_token = Token(TokenType.ASTERISK, self.char)
        elif self.char == "/":
            new_token = Token(TokenType.SLASH, self.char)
        elif self.char == "<":
            new_token = Token(TokenType.LT, self.char)
        elif self.char == ">":
            new_token = Token(TokenType.GT, self.char)
        elif self.char == "":
            new_token = Token(TokenType.EOF, "")

        # Two characters symbols
        elif self.char == "=":
            if self._peek_char() == "=":
                self._read_char()
                new_token = Token(TokenType.EQ, char + self.char)
            else:
                new_token = Token(TokenType.ASSIGN, self.char)
        elif self.char == "!":
            if self._peek_char() == "=":
                self._read_char()
                new_token = Token(TokenType.NOT_EQ, char + self.char)
            else:
                new_token = Token(TokenType.BANG, self.char)

        # Numbers and identifiers
        else:
            if self.char.isalpha():
                literal: str = self._read_identifier()
                new_token = Token(lookup_identifier(literal), literal)
                return new_token
            elif self.char.isdigit():
                new_token = Token(TokenType.INT, self._read_number())
                return new_token
            else:
                new_token = Token(TokenType.ILLEGAL, self.char)

        self._read_char()
        return new_token

    def _read_char(self) -> None:
        """
        This method is for traversing the source code
        character by character
        """
        if self.peek_position >= len(self.source):
            self.char = ""
        else:
            self.char = self.source[self.peek_position]
        self.position = self.peek_position
        self.peek_position += 1

    def _read_identifier(self) -> str:
        """
        This helper method extracts any indentifier from
        the source code like variable or function names
        """
        position: int = self.position
        while self.char.isalpha():
            self._read_char()
        return self.source[position : self.position]

    def _read_number(self) -> str:
        """
        This helper method extracts integers from the
        source code
        """
        position: int = self.position
        while self.char.isdigit():
            self._read_char()
        return self.source[position : self.position]

    def _skip_whitespace(self) -> None:
        """
        This helper method advances the position until
        no whitespace is encountered
        """
        while self.char.isspace():
            self._read_char()

    def _peek_char(self) -> str:
        """
        This helper method is somewhat like _read_char()
        without the increment, it's for seeing what's next,
        not to move around
        """
        if self.peek_position >= len(self.source):
            return ""
        return self.source[self.peek_position]



from __future__ import annotations
from enum import Enum
from typing import Callable
from lalalang.lexer.lexer import Lexer
from lalalang.lexer.token import Token, TokenType
from lalalang.parser.precedence import PRECEDENCES, ExpressionPrecedence
from lalalang.parser.ast import *


class Parser:
    def __init__(self, lexer: Lexer):
        self.lexer: Lexer = lexer
        self.current_token: Token = Token.empty()
        self.peek_token: Token = Token.empty()
        self.statements: list[Statement] = []
        self.errors: list[str] = []
        self.prefix_parse_funs: dict[TokenType, Callable] = dict({})
        self.infix_parse_funs: dict[TokenType, Callable] = dict({})
        self._setup()

    def __repr__(self):
        conversion: list[str] = [repr(i) for i in self.statements]
        return "\n".join(conversion)

    def __str__(self):
        conversion: list[str] = [str(i) for i in self.statements]
        return "".join(conversion)

    def parse_program(self) -> Program:
        """
        This is the main method that will traverse the tokens and generate
        the respective abstract syntax tree
        """
        program: Program = Program()
        while not self._current_token_is(TokenType.EOF):
            statement: Statement = self._parse_statement()
            if statement:
                program.add_statement(statement)
            self._next_token()
        return program

    def _setup(self) -> None:
        """Some stuff that needs to be ready at the construction of this class"""
        self._register_prefix_fun(TokenType.IDENT, self._parse_identifier)
        self._register_prefix_fun(TokenType.INT, self._parse_integer_literal)
        self._register_prefix_fun(TokenType.BANG, self._parse_prefix_expression)
        self._register_prefix_fun(TokenType.MINUS, self._parse_prefix_expression)
        self._register_prefix_fun(TokenType.TRUE, self._parse_boolean)
        self._register_prefix_fun(TokenType.FALSE, self._parse_boolean)
        self._register_prefix_fun(TokenType.LPAREN, self._parse_grouped_expression)
        self._register_prefix_fun(TokenType.IF, self._parse_if_expression)
        self._register_prefix_fun(TokenType.FUNCTION, self._parse_function_literal)

        self._register_infix_fun(TokenType.PLUS, self._parse_infix_expression)
        self._register_infix_fun(TokenType.MINUS, self._parse_infix_expression)
        self._register_infix_fun(TokenType.SLASH, self._parse_infix_expression)
        self._register_infix_fun(TokenType.ASTERISK, self._parse_infix_expression)
        self._register_infix_fun(TokenType.EQ, self._parse_infix_expression)
        self._register_infix_fun(TokenType.NOT_EQ, self._parse_infix_expression)
        self._register_infix_fun(TokenType.LT, self._parse_infix_expression)
        self._register_infix_fun(TokenType.GT, self._parse_infix_expression)
        self._register_infix_fun(TokenType.LPAREN, self._parse_call_expression)

        self._next_token()
        self._next_token()

    def _next_token(self) -> None:
        """This helps us moving arround the lexer to watch one token ahead"""
        self.current_token = self.peek_token
        self.peek_token = self.lexer.next_token()

    def _parse_statement(self) -> Statement:
        """Here are the conditionals to handle any statements defined by us"""
        if self._current_token_is(TokenType.LET):
            return self._parse_let_statement()
        elif self._current_token_is(TokenType.RETURN):
            return self._parse_return_statement()
        else:
            return self._parse_expression_statement()

    def _parse_let_statement(self) -> LetStatement:
        """Parse the let statement"""
        statement: LetStatement = LetStatement.empty()
        statement.token = self.current_token

        if not self._peek_expected(TokenType.IDENT):
            return None

        statement.name = Identifier(self.current_token, self.current_token.literal)

        if not self._peek_expected(TokenType.ASSIGN):
            return None

        self._next_token()
        statement.value = self._parse_expression(ExpressionPrecedence.LOWEST)

        if self._peek_token_is(TokenType.SEMICOLON):
            self._next_token()

        return statement

    def _parse_return_statement(self) -> ReturnStatement:
        """Pase the return statement"""
        statement: ReturnStatement = ReturnStatement.empty()
        statement.token = self.current_token

        self._next_token()

        statement.return_value = self._parse_expression(ExpressionPrecedence.LOWEST)
        if self._peek_token_is(TokenType.SEMICOLON):
            self._next_token()

        return statement

    def _parse_expression_statement(self) -> ExpressionStatement:
        """Parse the expression statement"""
        statement: ExpressionStatement = ExpressionStatement.empty()
        statement.expression = self._parse_expression(ExpressionPrecedence.LOWEST)

        if self._peek_token_is(TokenType.SEMICOLON):
            self._next_token()

        return statement

    def _parse_expression(self, precedence: ExpressionPrecedence) -> Expression:
        """
        This is the main method that contains explicitly pratt
        parsing, it is a recursive algorithm which decides about
        infix and prefix parsing functions.
        """
        prefix: Callable = self.prefix_parse_funs.get(self.current_token.token_type)
        if not prefix:
            self._no_prefix_parsing_error(self.current_token.token_type)
            return None

        left_expression: Expression = prefix()

        while (
            not self._peek_token_is(TokenType.SEMICOLON)
            and precedence < self._peek_precendence()
        ):
            infix: Callable = self.infix_parse_funs.get(self.peek_token.token_type)
            if not infix:
                return left_expression

            self._next_token()
            left_expression = infix(left_expression)

        return left_expression

    def _parse_identifier(self) -> Identifier:
        """Parse any identifier (e.g variable and function names)"""
        return Identifier(self.current_token, self.current_token.literal)

    def _parse_integer_literal(self) -> IntegerLiteral:
        """Parse an integer literal (the actual number)"""
        literal: IntegerLiteral = IntegerLiteral.empty()
        literal.token = self.current_token
        try:
            number: int = int(self.current_token.literal)
        except ValueError:
            self.errors.append(
                "Could not parse %s as integer" % self.current_token.literal
            )
        literal.value = number
        return literal

    def _parse_boolean(self) -> BooleanLiteral:
        """Parse boolean expression"""
        return BooleanLiteral(
            self.current_token, self._current_token_is(TokenType.TRUE)
        )

    def _parse_grouped_expression(self) -> Expression:
        """Parse any grouped expression (any expression with more than one operator)"""
        self._next_token()
        expression: Expression = self._parse_expression(ExpressionPrecedence.LOWEST)
        if self._peek_expected(TokenType.RPAREN):
            return expression
        return None

    def _parse_if_expression(self) -> IfExpression:
        """Parse if expressions"""
        expression: IfExpression = IfExpression.empty()
        expression.token = self.current_token

        if not self._peek_expected(TokenType.LPAREN):
            return None

        self._next_token()
        expression.condition = self._parse_expression(ExpressionPrecedence.LOWEST)

        if not self._peek_expected(TokenType.RPAREN):
            return None

        if not self._peek_expected(TokenType.LBRACE):
            return None

        expression.consequence = self._parse_block_statement()

        if self._peek_token_is(TokenType.ELSE):
            self._next_token()

            if not self._peek_expected(TokenType.LBRACE):
                return None

            expression.alternative = self._parse_block_statement()

        return expression

    def _parse_block_statement(self) -> BlockStatement:
        """Parse the block statements (consequent and alternatives of if expressions)"""
        block: BlockStatement = BlockStatement.empty()
        block.statements = []

        self._next_token()

        while not self._current_token_is(
            TokenType.RBRACE
        ) and not self._current_token_is(TokenType.EOF):
            statement = self._parse_statement()

            if statement:
                block.statements.append(statement)

            self._next_token()

        return block

    def _parse_function_literal(self) -> FunctionLiteral:
        """Parse the name, parameters and body of a function"""
        literal: FunctionLiteral = FunctionLiteral.empty()
        literal.token = self.current_token

        if not self._peek_expected(TokenType.LPAREN):
            return None

        literal.parameters = self._parse_function_parameters()

        if not self._peek_expected(TokenType.LBRACE):
            return None

        literal.body = self._parse_block_statement()
        return literal

    def _parse_function_parameters(self) -> list[Identifier]:
        """Parse the function parameters"""
        identifiers: list[Identifier] = []

        if self._peek_token_is(TokenType.RPAREN):
            self._next_token()
            return identifiers

        self._next_token()

        ident: Identifier = Identifier(self.current_token, self.current_token.literal)
        identifiers.append(ident)

        while self._peek_token_is(TokenType.COMMA):
            self._next_token()
            self._next_token()
            ident = Identifier(self.current_token, self.current_token.literal)
            identifiers.append(ident)

        if not self._peek_expected(TokenType.RPAREN):
            return None

        return identifiers

    def _parse_call_expression(self, function: Expression) -> CallExpression:
        """Parse a call expression (function call)"""
        expression: CallExpression = CallExpression.empty()
        expression.function = function
        expression.token = self.current_token
        expression.arguments = self._parse_call_arguments()
        return expression

    def _parse_call_arguments(self) -> list[Expression]:
        """Parse the call arguments"""
        arguments: list[Expression] = []

        if self._peek_token_is(TokenType.RPAREN):
            self._next_token()
            return arguments

        self._next_token()
        arguments.append(self._parse_expression(ExpressionPrecedence.LOWEST))

        while self._peek_token_is(TokenType.COMMA):
            self._next_token()
            self._next_token()
            arguments.append(self._parse_expression(ExpressionPrecedence.LOWEST))

        if not self._peek_expected(TokenType.RPAREN):
            return None

        return arguments

    def _parse_prefix_expression(self) -> PrefixExpression:
        """We return the prefix expression of our current token"""
        expression: PrefixExpression = PrefixExpression.empty()
        expression.token = self.current_token
        expression.operator = self.current_token.literal

        self._next_token()
        expression.right = self._parse_expression(ExpressionPrecedence.PREFIX)

        return expression

    def _parse_infix_expression(self, left: Expression) -> InfixExpression:
        """We return the infix expression of our current token"""
        expression: InfixExpression = InfixExpression.empty()
        expression.token = self.current_token
        expression.operator = self.current_token.literal
        expression.left = left

        precedence: ExpressionPrecedence = self._current_precedence()
        self._next_token()
        expression.right = self._parse_expression(precedence)

        return expression

    def _no_prefix_parsing_error(self, token_type: TokenType) -> None:
        """Add this error message whenever a prefix expression is unrecognized"""
        message = "No prefix function to parse %s" % token_type
        self.errors.append(message)

    def _current_precedence(self) -> ExpressionPrecedence:
        """Gets the precendence of the current token"""
        precendence: ExpressionPrecedence = PRECEDENCES.get(
            self.current_token.token_type
        )
        if not precendence:
            return ExpressionPrecedence.LOWEST
        return precendence

    def _current_token_is(self, token_type: TokenType) -> bool:
        """Check if the current token is a specific TokenType"""
        return self.current_token.token_type == token_type

    def _peek_token_is(self, token_type: TokenType) -> bool:
        """Check if the next token is a specific TokenType"""
        return self.peek_token.token_type == token_type

    def _peek_error(self, token_type: TokenType) -> None:
        """Add error message of unexpected tokens"""
        message: str = "Expected type %s, got %s" % (
            token_type,
            self.peek_token.token_type,
        )
        self.errors.append(message)

    def _peek_precendence(self) -> ExpressionPrecedence:
        """Gets the precedence of the peek token"""
        precedence: ExpressionPrecedence = PRECEDENCES.get(
            self.peek_token.token_type, ExpressionPrecedence.LOWEST
        )
        return precedence

    def _peek_expected(self, token_type: TokenType) -> bool:
        """This helper method is for checking if next token has the expected TokenType"""
        if self._peek_token_is(token_type):
            self._next_token()
            return True
        self._peek_error(token_type)
        return False

    def _register_prefix_fun(self, token_type: TokenType, fun: Callable) -> None:
        """Associate a token type with a function for the prefix statements"""
        self.prefix_parse_funs[token_type] = fun

    def _register_infix_fun(self, token_type: TokenType, fun: Callable) -> None:
        """Associate a token type with a function for the infix statements"""
        self.infix_parse_funs[token_type] = fun



from lalalang.lexer.token import TokenType, Token
from lalalang.parser.ast import Program, LetStatement, ReturnStatement, Identifier


CONSTRUCTED_PROGRAM = Program.with_statements(
    [
        LetStatement(
            token=Token(TokenType.LET, "let"),
            name=Identifier(token=Token(TokenType.IDENT, "myVar"), value="myVar"),
            value=Identifier(
                token=Token(TokenType.IDENT, "anotherVar"), value="anotherVar"
            ),
        )
    ]
)

LET_STATEMENTS = [
    {
        "input": "let xyz = 412;",
        "expected": ["xyz", "412"],
    },
    {
        "input": "let variable = value;",
        "expected": ["variable", "value"],
    },
    {
        "input": "let i = 0;",
        "expected": ["i", "0"],
    },
]

RETURN_STATEMENTS = [
    {
        "input": "return 5;",
        "expected": "5",
    },
    {
        "input": "return 1000;",
        "expected": "1000",
    },
    {
        "input": "return hello;",
        "expected": "hello",
    },
]

IDENT_EXPRESSION = [
    {
        "input": "another;",
        "expected": "another",
    },
    {
        "input": "day;",
        "expected": "day",
    },
    {
        "input": "of;",
        "expected": "of",
    },
    {
        "input": "sun;",
        "expected": "sun",
    },
]

INT_LITERALS = [
    {
        "input": "5;",
        "expected": 5,
    },
    {
        "input": "314;",
        "expected": 314,
    },
    {
        "input": "278;",
        "expected": 278,
    },
]

PREFIX_EXPRESSIONS = [
    {
        "input": "!5;",
        "expected": ["!", 5],
    },
    {
        "input": "-10;",
        "expected": ["-", 10],
    },
]

INFIX_EXPRESSIONS = [
    {
        "input": "5 + 5;",
        "expected": [5, "+", 5],
    },
    {
        "input": "5 - 5;",
        "expected": [5, "-", 5],
    },
    {
        "input": "5 * 5;",
        "expected": [5, "*", 5],
    },
    {
        "input": "5 / 5;",
        "expected": [5, "/", 5],
    },
    {
        "input": "5 > 5;",
        "expected": [5, ">", 5],
    },
    {
        "input": "5 < 5;",
        "expected": [5, "<", 5],
    },
    {
        "input": "5 == 5;",
        "expected": [5, "==", 5],
    },
    {
        "input": "5 != 5;",
        "expected": [5, "!=", 5],
    },
]

PRECEDENCE_EXPRESSIONS = [
    {
        "input": "-a * b",
        "expected": "((- a) * b)",
    },
    {
        "input": "!-a",
        "expected": "(! (- a))",
    },
    {
        "input": "a + b + c",
        "expected": "((a + b) + c)",
    },
    {
        "input": "a * b * c",
        "expected": "((a * b) * c)",
    },
    {
        "input": "a * b / c",
        "expected": "((a * b) / c)",
    },
    {
        "input": "a + b / c",
        "expected": "(a + (b / c))",
    },
    {
        "input": "a + b * c + d / e - f",
        "expected": "(((a + (b * c)) + (d / e)) - f)",
    },
    {
        "input": "3 + 4; -5 * 5",
        "expected": "(3 + 4)((- 5) * 5)",
    },
    {
        "input": "5 > 4 == 3 < 4",
        "expected": "((5 > 4) == (3 < 4))",
    },
    {
        "input": "5 < 4 != 3 > 4",
        "expected": "((5 < 4) != (3 > 4))",
    },
]

BOOLEAN_EXPRESSION = [
    {
        "input": "true",
        "expected": "true",
    },
    {
        "input": "false",
        "expected": "false",
    },
]

GROUPED_EXPRESSIONS = [
    {
        "input": "1 + (2 + 3) + 4",
        "expected": "((1 + (2 + 3)) + 4)",
    },
    {
        "input": "(5 + 5) * 2",
        "expected": "((5 + 5) * 2)",
    },
    {
        "input": "2 / (5 + 5)",
        "expected": "(2 / (5 + 5))",
    },
    {
        "input": "-(5 + 5)",
        "expected": "(- (5 + 5))",
    },
    {
        "input": "!(true == true)",
        "expected": "(! (true == true))",
    },
]

IF_EXPRESSIONS = [
    {
        "input": "if (x < y) { x }",
        "expected": ["(x < y)", "x", "None"],
    },
    {
        "input": "if (x > y) { x } else { y }",
        "expected": ["(x > y)", "x", "y"],
    },
]

FUNCTION_LITERALS = [
    {
        "input": "fun(x, y) { x + y; }",
        "expected": ["fun(x,y) (x + y)", "(x + y)"],
    },
    {
        "input": "fun(x) { x - x; }",
        "expected": ["fun(x) (x - x)", "(x - x)"],
    },
]

CALL_EXPRESSIONS = [
    {
        "input": "start(fire);",
        "expected": ["start", "start(fire)"],
    },
    {
        "input": "city(of,2,stars);",
        "expected": ["city", "city(of, 2, stars)"],
    },
    {
        "input": "planet(1,2);",
        "expected": ["planet", "planet(1, 2)"],
    },
]



from lalalang.parser.ast import *
from lalalang.evaluator.object import Object, Integer, Boolean, Null

# References
TRUE = Boolean(True)
FALSE = Boolean(False)
NULL = Null()


def eval_3lang(node: Node) -> Object:
    """
    This is executing a tree walk interpreter, doing a
    postorder traverse over the ast
    """

    # Statements
    if isinstance(node, Program):
        return eval_statements(node.statements)

    # Expressions
    elif isinstance(node, ExpressionStatement):
        return eval_3lang(node.expression)

    # Internal objects
    elif isinstance(node, IntegerLiteral):
        return Integer(node.value)
    elif isinstance(node, BooleanLiteral):
        return Boolean(node.value)

    return None


def eval_statements(statements: list[Statement]) -> Object:
    """
    The root of every program is a list of statements, not
    the ast nodes, we need to start from the root
    """
    for statement in statements:
        result: Object = eval_3lang(statement)
    return result



from __future__ import annotations
from lalalang.lexer.token import TokenType, Token
from lalalang.parser.nodes import Node, Statement, Expression


class Program(Node):
    """
    This class is the root node of every abstract
    syntax tree
    """

    @staticmethod
    def with_statements(statements: list[Statement]) -> Program:
        program: Program = Program()
        program.statements = statements
        return program

    def __init__(self):
        self.statements: list[Statement] = []

    def __repr__(self):
        conversion: list[str] = [repr(i) for i in self.statements]
        return "\n".join(conversion)

    def __str__(self):
        conversion: list[str] = [str(i) for i in self.statements]
        return "".join(conversion)

    def token_literal(self) -> str:
        if len(self.statements) > 0:
            return self.statements[0].token_literal()
        return ""

    def add_statement(self, statement: Statement):
        self.statements.append(statement)


class Identifier(Expression):
    """
    This is the class for dealing with the names or
    identifier of a let expression
    """

    @staticmethod
    def empty():
        return Identifier(None, None)

    def __init__(self, token: Token, value: str):
        self.token: Token = token
        self.value: str = value

    def __repr__(self):
        return "Identifier(%s, %s)" % (repr(self.token), self.value)

    def __str__(self):
        return self.value

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self) -> None:
        pass


class LetStatement(Statement):
    """
    This is for let statements, binding names or
    identifiers to values
    """

    @staticmethod
    def empty():
        return LetStatement(None, None, None)

    def __init__(self, token: Token, name: Identifier, value: Expression):
        self.token: Token = token
        self.name: Identifier = name
        self.value: Expression = value

    def __repr__(self):
        return "LetStatement(%s, %s, %s)" % (
            repr(self.token),
            repr(self.name),
            repr(self.value),
        )

    def __str__(self):
        if not self.value:
            return "%s %s = ;" % (self.token_literal(), str(self.name))
        return "%s %s = %s;" % (self.token_literal(), str(self.name), str(self.value))

    def token_literal(self) -> str:
        return self.token.literal

    def statement_node(self) -> None:
        pass


class ReturnStatement(Statement):
    """
    This is for return statements, for reusing
    values made inside functions
    """

    @staticmethod
    def empty():
        return ReturnStatement(None, None)

    def __init__(self, token: Token, return_value: Expression):
        self.token: Token = token
        self.return_value: Expression = return_value

    def __repr__(self):
        return "ReturnStatement(%s, %s)" % (repr(self.token), repr(self.return_value))

    def __str__(self):
        return "%s %s;" % (self.token_literal(), str(self.return_value))

    def token_literal(self) -> str:
        return self.token.literal

    def statement_node(self) -> None:
        pass


class ExpressionStatement(Statement):
    """
    This is for every line of code that returns a value,
    this is necessary to have simpler expressions to print
    the result
    """

    @staticmethod
    def empty():
        return ExpressionStatement(None, None)

    def __init__(self, token: Token, expression: Expression):
        self.token: Token = token
        self.expression: Expression = expression

    def __repr__(self):
        return "ExpressionStatement(%s, %s)" % (repr(self.token), repr(self.expression))

    def __str__(self):
        if not self.expression:
            return ""
        return str(self.expression)

    def token_literal(self) -> str:
        return self.token.literal

    def statement_node(self) -> None:
        pass


class IntegerLiteral(Expression):
    """
    This is for representing the actual numbers, it is
    a expression so that they can return themselves as the
    value represented.
    """

    @staticmethod
    def empty():
        return IntegerLiteral(None, None)

    def __init__(self, token: Token, value: int):
        self.token: Token = token
        self.value: int = value

    def __repr__(self):
        return "IntegerLiteral(%s,%s)" % (repr(self.token), self.value)

    def __str__(self):
        return str(self.value)

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self):
        pass


class PrefixExpression(Expression):
    """
    This is any expression that has an operator before the
    operand for now we only have two prefix operators - and !.
    """

    @staticmethod
    def empty():
        return PrefixExpression(None, None, None)

    def __init__(self, token: Token, operator: str, right: Expression):
        self.token: Token = token
        self.operator: str = operator
        self.right: Expression = right

    def __repr__(self):
        return "PrefixExpression(%s,%s,%s)" % (
            repr(self.token),
            self.operator,
            repr(self.right),
        )

    def __str__(self):
        return "(%s %s)" % (self.operator, str(self.right))

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self):
        pass


class InfixExpression(Expression):
    """
    This is any expression that has an operator sitting between
    its operands, here are most arithmetic operations
    """

    @staticmethod
    def empty():
        return InfixExpression(None, None, None, None)

    def __init__(
        self, token: Token, left: Expression, operator: str, right: Expression
    ):
        self.token: Token = token
        self.left: Expression = left
        self.operator: str = operator
        self.right: Expression = right

    def __repr__(self):
        return "InfixExpression(%s,%s,%s,%s)" % (
            repr(self.token),
            repr(self.left),
            repr(self.operator),
            repr(self.right),
        )

    def __str__(self):
        return "(%s %s %s)" % (str(self.left), self.operator, str(self.right))

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self):
        pass


class BooleanLiteral(Expression):
    """
    Here we only store the simplest for of data, a boolean they
    are represented with the literals 'true' or 'false'
    """

    def __init__(self, token: Token, value: bool):
        self.token: Token = token
        self.value: bool = value

    def __repr__(self):
        return "BooleanLiteral(%s, %s)" % (repr(self.token), self.value)

    def __str__(self):
        return str(self.token.literal)

    def token_literal(self):
        return self.token.literal

    def expression_node(self):
        pass


class IfExpression(Expression):
    """
    This are conditionals which has an expression that evaluates to
    a boolean (also known as predicate), a consequence or alternative
    that executes depending the condition
    """

    @staticmethod
    def empty():
        return IfExpression(None, None, None, None)

    def __init__(
        self,
        token: Token,
        condition: Expression,
        consequence: BlockStatement,
        alternative: BlockStatement,
    ):
        self.token: Token = token
        self.condition: Expression = condition
        self.consequence: BlockStatement = consequence
        self.alternative: BlockStatement = alternative

    def __repr__(self):
        return "IfExpression(%s, %s, %s, %s)" % (
            repr(self.token),
            repr(self.condition),
            repr(self.consequence),
            repr(self.alternative),
        )

    def __str__(self):
        if self.alternative:
            return "if %s %s else %s" % (
                str(self.condition),
                str(self.consequence),
                str(self.alternative),
            )
        return "if %s %s" % (str(self.condition), str(self.consequence))

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self):
        pass


class BlockStatement(Statement):
    """
    This is a construct for storing various statements that for
    example is the body of an if expression or function literal
    """

    @staticmethod
    def empty():
        return BlockStatement(None, None)

    def __init__(self, token: Token, statements: list[Statement]):
        self.token: Token = token
        self.statements: list[Statement] = statements

    def __repr__(self):
        conv_stmnts = [repr(i) for i in self.statements]
        return "BlockStatement(%s, %s)" % (str(self.token), "".join(conv_stmnts))

    def __str__(self):
        conv_stmnts = [str(i) for i in self.statements]
        return "".join(conv_stmnts)

    def token_literal(self) -> str:
        return self.token.literal

    def statement_node(self):
        pass


class FunctionLiteral(Expression):
    """
    This is fore storing the a function, with its name and parameters
    (also known as signature) and the body, which is a BlockStatement
    """

    @staticmethod
    def empty():
        return FunctionLiteral(None, None, None)

    def __init__(
        self, token: Token, parameters: list[Identifier], body: BlockStatement
    ):
        self.token: Token = token
        self.parameters: list[Identifier] = parameters
        self.body: BlockStatement = body

    def __repr__(self):
        convs_params = [repr(i) for i in self.parameters]
        return "FunctionLiteral(%s, %s, %s)" % (
            repr(self.token),
            "".join(convs_params),
            repr(self.body),
        )

    def __str__(self):
        convs_params = [str(i) for i in self.parameters]
        return "%s(%s) %s" % (
            self.token_literal(),
            ",".join(convs_params),
            str(self.body),
        )

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self):
        pass


class CallExpression(Expression):
    """
    Here are the expressions that are using our declared function
    literals, we only store a the funcion used with its arguments
    """

    @staticmethod
    def empty():
        return CallExpression(None, None, None)

    def __init__(self, token: Token, function: Expression, arguments: list[Expression]):
        self.token: Token = token
        self.function: Expression = function
        self.arguments: list[Expression] = arguments

    def __repr__(self):
        convs_args = [repr(i) for i in self.arguments]
        return "CallExpression(%s, %s, %s)" % (
            repr(self.token),
            repr(self.function),
            ", ".join(convs_args),
        )

    def __str__(self):
        convs_args = [str(i) for i in self.arguments]
        return "%s(%s)" % (str(self.function), ", ".join(convs_args))

    def token_literal(self) -> str:
        return self.token.literal

    def expression_node(self):
        pass



from abc import ABC, abstractmethod


class Node(ABC):
    """
    This class represents every node that is
    part of the abstract syntax tree
    """

    @abstractmethod
    def token_literal(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class Statement(Node):
    """
    These nodes are reserved only for the
    statements, meaning it won't produce any value.
    In la la lang there there are only two staments
    return and let.
    """

    @abstractmethod
    def statement_node(self) -> None:
        pass


class Expression(Node):
    """
    These nodes are reserved only for the
    expression, meaning it will produce a value. Aside
    from let and return everything is a expression.
    """

    @abstractmethod
    def expression_node(self) -> None:
        pass



from enum import Enum


class TokenType(Enum):
    """
    This are all the token types available in an
    enumeration
    """

    def __repr__(self):
        return str(self.name)

    def __str__(self):
        return str(self.value)

    # Control
    ILLEGAL = "ILLEGAL"
    EOF = "EOF"

    # Identifiers and literals
    IDENT = "IDENT"
    INT = "INT"

    # Operators
    ASSIGN = "="
    PLUS = "+"
    MINUS = "-"
    BANG = "!"
    ASTERISK = "*"
    SLASH = "/"
    LT = "<"
    GT = ">"
    EQ = "=="
    NOT_EQ = "!="

    # Delimeters
    COMMA = ","
    SEMICOLON = ";"
    LPAREN = "("
    RPAREN = ")"
    LBRACE = "{"
    RBRACE = "}"

    # Keywords
    FUNCTION = "FUNCTION"
    LET = "LET"
    TRUE = "TRUE"
    FALSE = "FALSE"
    IF = "IF"
    ELSE = "ELSE"
    RETURN = "RETURN"


class Token:
    """
    This class is for representing the object the
    lexer is producing
    """

    @staticmethod
    def empty():
        return Token(None, None)

    def __init__(self, token_type: TokenType, literal: str):
        self.token_type: TokenType = token_type
        self.literal: str = literal

    def __repr__(self):
        return "Token(%s, %s)" % (repr(self.token_type), self.literal)

    def __str__(self):
        return "Token(%s, %s)" % (self.token_type, self.literal)


KEYWORDS: dict[str, TokenType] = {
    "fun": TokenType.FUNCTION,
    "let": TokenType.LET,
    "true": TokenType.TRUE,
    "false": TokenType.FALSE,
    "if": TokenType.IF,
    "else": TokenType.ELSE,
    "return": TokenType.RETURN,
}


def lookup_identifier(key: str) -> TokenType:
    """
    This function is for getting the token type of an letters-name,
    so we can't confuse between keywords and identifiers
    """
    if key in KEYWORDS:
        return KEYWORDS[key]
    return TokenType.IDENT



from __future__ import annotations
from abc import ABC, abstractmethod
from enum import Enum


class Object(ABC):
    """
    This is how we internally represent the objects in the
    interpreter in order to manipulate it
    """

    @abstractmethod
    def object_type(self) -> ObjectType:
        pass

    @abstractmethod
    def inspect(self) -> str:
        pass


class ObjectType(Enum):
    """These are all the possible objects we can have"""

    INTEGER = "INTEGER"
    BOOLEAN = "BOOLEAN"
    NULL = "NULL"


class Integer(Object):
    def __init__(self, value: int):
        self.value: int = value

    def object_type(self) -> ObjectType:
        return ObjectType.INTEGER

    def inspect(self) -> str:
        return str(self.value)


class Boolean(Object):
    def __init__(self, value: bool):
        self.value = value

    def object_type(self) -> ObjectType:
        return ObjectType.BOOLEAN

    def inspect(self) -> str:
        return str(self.value)


class Null(Object):
    def object_type(self) -> ObjectType:
        return ObjectType.NULL

    def inspect(self) -> str:
        return "null"



from unittest import TestCase
from lalalang.lexer.lexer import Lexer
from lalalang.lexer.token import TokenType
from lalalang.parser.parser import Parser
from lalalang.parser.ast import *
from tests.mocks.parser import *


class TestParserStatements(TestCase):
    def test_program_construction(self):
        result = "let myVar = anotherVar;"
        self.assertEqual(str(CONSTRUCTED_PROGRAM), result)

    def test_let_statements(self):
        for ls in LET_STATEMENTS:
            source = ls.get("input")
            program = self.create_program(source)

            output = ls.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, LetStatement)

            expression = statement.value
            self.assertIsInstance(expression, Expression)

            self.assertEqual(str(statement.name), output[0])
            self.assertEqual(str(expression), output[1])

    def test_return_statements(self):
        for rs in RETURN_STATEMENTS:
            source = rs.get("input")
            program = self.create_program(source)

            output = rs.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ReturnStatement)

            expression = statement.return_value
            self.assertEqual(str(expression), output)

    def test_identifiers_expressions(self):
        for ie in IDENT_EXPRESSION:
            source = ie.get("input")
            program = self.create_program(source)

            output = ie.get("expected")
            self.assertEqual(str(program), output)
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, Identifier)

    def test_int_literal(self):
        for il in INT_LITERALS:
            source = il.get("input")
            program = self.create_program(source)

            output = il.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertTrue(self.validate_integer_literal(expression, output))

    def test_prefix_expressions(self):
        for pe in PREFIX_EXPRESSIONS:
            source = pe.get("input")
            program = self.create_program(source)

            output = pe.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, PrefixExpression)
            self.assertEqual(expression.operator, output[0])
            self.assertTrue(self.validate_integer_literal(expression.right, output[1]))

    def test_infix_expressions(self):
        for ie in INFIX_EXPRESSIONS:
            source = ie.get("input")
            program = self.create_program(source)

            output = ie.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, InfixExpression)
            self.assertEqual(expression.operator, output[1])

            self.assertTrue(self.validate_integer_literal(expression.left, output[0]))
            self.assertEqual(expression.operator, output[1])
            self.assertTrue(self.validate_integer_literal(expression.right, output[2]))

    def test_precedence(self):
        for pe in PRECEDENCE_EXPRESSIONS:
            source = pe.get("input")
            program = self.create_program(source)

            expected = pe.get("expected")
            self.assertEqual(str(program), expected)

    def test_boolean_expressions(self):
        for be in BOOLEAN_EXPRESSION:
            source = be.get("input")
            program = self.create_program(source)

            expected = be.get("expected")
            self.assertEqual(str(program), expected)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, BooleanLiteral)

    def test_grouped_expressions(self):
        for ge in GROUPED_EXPRESSIONS:
            source = ge.get("input")
            program = self.create_program(source)

            expected = ge.get("expected")
            self.assertEqual(str(program), expected)

    def test_if_expressions(self):
        for ie in IF_EXPRESSIONS:
            source = ie.get("input")
            program = self.create_program(source)

            expected = ie.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, IfExpression)

            consequence = expression.consequence
            self.assertIsInstance(consequence, BlockStatement)

            self.assertEqual(str(expression.condition), expected[0])
            self.assertEqual(str(expression.consequence), expected[1])
            self.assertEqual(str(expression.alternative), expected[2])

    def test_function_literals(self):
        for fl in FUNCTION_LITERALS:
            source = fl.get("input")
            program = self.create_program(source)

            expected = fl.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, FunctionLiteral)

            body = expression.body
            self.assertIsInstance(body, BlockStatement)

            self.assertEqual(str(expression), expected[0])
            self.assertEqual(str(body), expected[1])

    def test_call_expressions(self):
        for ce in CALL_EXPRESSIONS:
            source = ce.get("input")
            program = self.create_program(source)

            expected = ce.get("expected")
            self.assertEqual(len(program.statements), 1)

            statement = program.statements[0]
            self.assertIsInstance(statement, ExpressionStatement)

            expression = statement.expression
            self.assertIsInstance(expression, CallExpression)

            self.assertEqual(str(expression.function), expected[0])
            self.assertEqual(str(expression), expected[1])

    def create_program(self, source_code):
        lex = Lexer(source_code)
        par = Parser(lex)
        return par.parse_program()

    def validate_integer_literal(self, integer, value):
        correct_type = isinstance(integer, IntegerLiteral)
        correct_value = integer.value == value
        correct_literal = integer.token_literal() == "%s" % value
        return correct_type and correct_value and correct_literal



import sys
import click
import signal
from lalalang.parser.parser import Parser
from lalalang.lexer.lexer import Lexer
from lalalang.lexer.token import TokenType, Token
from lalalang.evaluator.evaluator import eval_3lang


@click.command()
@click.option("--mode", default="eval", help="REPL mode (lex|parse|eval)")
@click.option("--src", default=None, help="Input file with 3lang code")
def cli(mode, src):

    code = None
    if not src:
        print("Welcome to the city of stars!🌟")
        print("This is the La La Lang Programming Languag v0.3!")
    else:
        with open(src, "r") as f:
            code = f.read().replace("\n", "")

    if mode == "lex":
        repl(lexing, code)
    elif mode == "parse":
        repl(parsing, code)
    elif mode == "eval":
        repl(evaluating, code)


def lexing(code):
    lex = Lexer(code)
    token = Token.empty()
    while token.token_type != TokenType.EOF:
        token = lex.next_token()
        print(repr(token))


def parsing(code):
    lex = Lexer(code)
    par = Parser(lex)
    program = par.parse_program()

    if len(par.errors) == 0:
        print(program)
    else:
        [print("\t %s" % err) for err in par.errors]


def evaluating(code):
    lex = Lexer(code)
    par = Parser(lex)
    program = par.parse_program()

    if len(par.errors) > 0:
        [print(i) for i in par.errors]
        return

    evaluated = eval_3lang(program)
    if evaluated:
        print(evaluated.inspect())


def repl(fun, source):
    if not source:
        while line := input("♪♪ > "):
            fun(line)
    fun(source)


def ctrlc_handler(sig, fram):
    print("\nSee you at Seb's 😉")
    sys.exit(0)


signal.signal(signal.SIGINT, ctrlc_handler)


if __name__ == "__main__":
    cli()



from unittest import TestCase
from lalalang.lexer.lexer import Lexer
from lalalang.lexer.token import TokenType, Token
from tests.mocks.lexer import *


class TestLexer(TestCase):
    def test_lexer_with_source_code(self):
        expected_tokens = [
            Token(TokenType.LET, "let"),
            Token(TokenType.IDENT, "pi"),
            Token(TokenType.ASSIGN, "="),
            Token(TokenType.INT, "314"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.FUNCTION, "fun"),
            Token(TokenType.IDENT, "area"),
            Token(TokenType.LPAREN, "("),
            Token(TokenType.IDENT, "ratio"),
            Token(TokenType.RPAREN, ")"),
            Token(TokenType.LBRACE, "{"),
            Token(TokenType.IDENT, "pi"),
            Token(TokenType.ASTERISK, "*"),
            Token(TokenType.IDENT, "ratio"),
            Token(TokenType.ASTERISK, "*"),
            Token(TokenType.IDENT, "ratio"),
            Token(TokenType.PLUS, "+"),
            Token(TokenType.INT, "0"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.RBRACE, "}"),
            Token(TokenType.INT, "13"),
            Token(TokenType.SLASH, "/"),
            Token(TokenType.INT, "3"),
            Token(TokenType.GT, ">"),
            Token(TokenType.INT, "16"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.IF, "if"),
            Token(TokenType.LPAREN, "("),
            Token(TokenType.INT, "10"),
            Token(TokenType.MINUS, "-"),
            Token(TokenType.INT, "8"),
            Token(TokenType.LT, "<"),
            Token(TokenType.INT, "16"),
            Token(TokenType.RPAREN, ")"),
            Token(TokenType.LBRACE, "{"),
            Token(TokenType.RETURN, "return"),
            Token(TokenType.TRUE, "true"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.RBRACE, "}"),
            Token(TokenType.ELSE, "else"),
            Token(TokenType.LBRACE, "{"),
            Token(TokenType.RETURN, "return"),
            Token(TokenType.FALSE, "false"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.RBRACE, "}"),
            Token(TokenType.INT, "10"),
            Token(TokenType.EQ, "=="),
            Token(TokenType.INT, "10"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.INT, "10"),
            Token(TokenType.NOT_EQ, "!="),
            Token(TokenType.INT, "9"),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.EOF, ""),
        ]
        self.compare_results(SOURCE_CODE, expected_tokens)

    def test_delimeters(self):
        expected_tokens = [
            Token(TokenType.LPAREN, "("),
            Token(TokenType.RPAREN, ")"),
            Token(TokenType.LBRACE, "{"),
            Token(TokenType.RBRACE, "}"),
            Token(TokenType.COMMA, ","),
            Token(TokenType.SEMICOLON, ";"),
            Token(TokenType.EOF, ""),
        ]
        self.compare_results(DELIMETERS, expected_tokens)

    def test_illegal(self):
        expected_tokens = [
            Token(TokenType.ILLEGAL, "~"),
            Token(TokenType.ILLEGAL, "`"),
            Token(TokenType.ILLEGAL, "^"),
            Token(TokenType.EOF, ""),
        ]
        self.compare_results(ILLEGAL, expected_tokens)

    def test_operators(self):
        expected_tokens = [
            Token(TokenType.PLUS, "+"),
            Token(TokenType.ASSIGN, "="),
            Token(TokenType.MINUS, "-"),
            Token(TokenType.BANG, "!"),
            Token(TokenType.ASTERISK, "*"),
            Token(TokenType.SLASH, "/"),
            Token(TokenType.LT, "<"),
            Token(TokenType.GT, ">"),
        ]
        self.compare_results(OPERATORS, expected_tokens)

    def test_two_character_symbols(self):
        expected_tokens = [
            Token(TokenType.BANG, "!"),
            Token(TokenType.EQ, "=="),
            Token(TokenType.NOT_EQ, "!="),
            Token(TokenType.ASSIGN, "="),
        ]
        self.compare_results(TWO_CHARACTER_SYMBOLS, expected_tokens)

    def test_identifiers(self):
        expected_tokens = [
            Token(TokenType.IDENT, "someone"),
            Token(TokenType.IDENT, "in"),
            Token(TokenType.IDENT, "the"),
            Token(TokenType.IDENT, "crowd"),
        ]
        self.compare_results(IDENTIFIERS, expected_tokens)

    def test_numbers(self):
        expected_tokens = [
            Token(TokenType.INT, "314"),
            Token(TokenType.INT, "217"),
            Token(TokenType.INT, "161"),
        ]
        self.compare_results(NUMBERS, expected_tokens)

    def test_keywords(self):
        expected_tokens = [
            Token(TokenType.LET, "let"),
            Token(TokenType.FUNCTION, "fun"),
            Token(TokenType.TRUE, "true"),
            Token(TokenType.FALSE, "false"),
            Token(TokenType.IF, "if"),
            Token(TokenType.ELSE, "else"),
            Token(TokenType.RETURN, "return"),
        ]
        self.compare_results(KEYWORDS, expected_tokens)

    def compare_results(self, source_code, expected_tokens):
        lex = Lexer(source_code)
        for token in expected_tokens:
            current = lex.next_token()
            self.assertEqual(token.token_type, current.token_type)
            self.assertEqual(token.literal, current.literal)
